#
# SPDX-License-Identifier: MIT
#
# Copyright (c) 2021-2023 Carsten Igel.
#
# This file is part of pdm-bump
# (see https://github.com/carstencodes/pdm-bump).
#
# This file is published using the MIT license.
# Refer to LICENSE for more information
#

""""""


from ..core.loader import loader as _loader
from .core import (
    DefaultVcsProvider,
    VcsProvider,
    VcsProviderAggregator,
    VcsProviderFactory,
    VcsProviderRegistry,
    vcs_provider,
    vcs_providers,
)

# Justification: Add items to registry
_loader.load_modules_of_package(
    __file__, __name__, "__init__", "core", "mixins"
)

__all__ = (
    "vcs_provider",
    "vcs_providers",
    "VcsProvider",
    "VcsProviderAggregator",
    "VcsProviderFactory",
    "VcsProviderRegistry",
    "DefaultVcsProvider",
)



#
# SPDX-License-Identifier: MIT
#
# Copyright (c) 2021-2023 Carsten Igel.
#
# This file is part of pdm-bump
# (see https://github.com/carstencodes/pdm-bump).
#
# This file is published using the MIT license.
# Refer to LICENSE for more information
#

from typing import Optional, Generator
from collections.abc import Iterable
from argparse import ArgumentParser, Namespace, ArgumentError

from pdm.project.config import ConfigItem
from pdm.cli.commands.base import BaseCommand

from pdm_bump.cli import main
from pdm_bump.plugin import BumpCommand

import pytest

parametrize = pytest.mark.parametrize
assert_raises = pytest.raises
fixture = pytest.fixture


_registered_configurations: dict[str, ConfigItem] = {}


class _CoreStub:
    registered_command: Optional[type[BaseCommand]] = None
    registered_name: Optional[str] = None

    def register_command(
        self, command: type[BaseCommand], name: Optional[str] = None
    ) -> None:
        self.registered_command = command
        self.registered_name = name

    @staticmethod
    def add_config(name: str, config_item: ConfigItem) -> None:
        _registered_configurations[str] = config_item

@fixture
def registrations() -> Generator:
    yield
    _registered_configurations.clear()

def test_registration_command_type(registrations) -> None:
    core: _CoreStub = _CoreStub()
    main(core)
    assert core.registered_command == BumpCommand, "The command is registered correctly"

def test_registration_command_name(registrations) -> None:
    core: _CoreStub = _CoreStub()
    main(core)
    assert core.registered_name is None, "No name has been supplied"

def test_registered_config_items(registrations) -> None:
    core: _CoreStub = _CoreStub()
    main(core)
    assert len(_registered_configurations) == 0, "No config item has been registered"


SUB_TEST_PARAMS_WHAT_SUCCESS: Iterable[tuple[str, str]] = [
    ("major", "major"),
    ("minor", "minor"),
    ("micro", "micro"),
    ("patch", "patch"),
    ("pre-release", "pre-release"),
    ("no-pre-release", "no-pre-release"),
    ("post", "post"),
    ("dev", "dev"),
]
SUB_TEST_PARAMS_PRE_SUCCESS: Iterable[tuple[str, str]] = [
    ("alpha", "alpha"),
    ("beta", "beta"),
    ("c", "c"),
    ("rc", "rc"),
]
SUB_TEST_PARAMS_FAIL: Iterable[tuple[str, list[str]]] = [
    ("major", ["maj"]),
    ("minor", ["min"]),
    ("micro", ["mic"]),
    ("patch", ["pa"]),
    ("pre-release", ["pre"]),
    ("no-pre-release", ["no-pre"]),
    ("alpha", ["pre-release", "--pre", "a"]),
    ("beta", ["pre-release", "--pre", "b"]),
    ("c", ["pre-release", "--pre", "release"]),
    ("rc", ["pre-release", "--pre", "release-candidate"]),
]

@parametrize(",".join(["msg", "what_success"]), SUB_TEST_PARAMS_WHAT_SUCCESS)
def test_parser_setup_what_success(msg, what_success) -> None:
    parser: ArgumentParser = ArgumentParser()
    _ = BumpCommand.init_parser(parser)

    options: Namespace = parser.parse_args([what_success])

    assert options.selected_command == what_success, "The first item is only matched"

@parametrize(",".join(["msg", "pre_success"]), SUB_TEST_PARAMS_PRE_SUCCESS)
def test_parser_setup_pre_success(msg, pre_success) -> None:
    parser: ArgumentParser = ArgumentParser()
    _ = BumpCommand.init_parser(parser)

    options: Namespace = parser.parse_args(
        ["pre-release", "--pre", pre_success]
    )

    assert options.pre_release_part == pre_success, "The second item is only matched"

@parametrize(",".join(["msg", "options"]), SUB_TEST_PARAMS_FAIL)
def test_parser_setup_fail(msg, options) -> None:
    parser: ArgumentParser = ArgumentParser(exit_on_error=False)
    _ = BumpCommand.init_parser(parser)

    with assert_raises(ArgumentError):
        parser.parse_args(options)



#
# SPDX-License-Identifier: MIT
#
# Copyright (c) 2021-2023 Carsten Igel.
#
# This file is part of pdm-bump
# (see https://github.com/carstencodes/pdm-bump).
#
# This file is published using the MIT license.
# Refer to LICENSE for more information
#
""""""


from abc import abstractmethod
from collections.abc import Iterator
from pathlib import Path

from .core import VcsFileSystemIdentifier, VcsProvider, VcsProviderFactory


class GitCommonVcsProviderFactory(VcsProviderFactory):
    """"""

    __git_dir_provider = VcsFileSystemIdentifier(
        file_name=None, dir_name=".git"
    )
    __git_wt_file_provider = VcsFileSystemIdentifier(
        file_name=".git", dir_name=None
    )

    @abstractmethod
    def _create_provider(self, path: Path) -> VcsProvider:
        """

        Parameters
        ----------
        path: Path :


        Returns
        -------

        """
        raise NotImplementedError()

    @property
    def vcs_fs_root(self) -> Iterator[VcsFileSystemIdentifier]:
        """"""
        yield self.__git_dir_provider
        yield self.__git_wt_file_provider



#
# SPDX-License-Identifier: MIT
#
# Copyright (c) 2021-2023 Carsten Igel.
#
# This file is part of pdm-bump
# (see https://github.com/carstencodes/pdm-bump).
#
# This file is published using the MIT license.
# Refer to LICENSE for more information
#
""""""


from abc import ABC, abstractmethod
from argparse import ArgumentParser, Namespace
from typing import Any, Callable, Protocol

from ..core.logging import logger
from ..core.version import Pep440VersionFormatter, Version
from ..vcs.core import VcsProvider

_formatter = Pep440VersionFormatter()


# Justification: Just a protocol
class _HasAddSubParser(Protocol):  # pylint: disable=R0903
    """"""

    def add_parser(self, name, **kwargs) -> ArgumentParser:
        """

        Parameters
        ----------
        name :

        **kwargs :


        Returns
        -------

        """
        raise NotImplementedError()


# Justification: Just a protocol
class VersionPersister(Protocol):  # pylint: disable=R0903
    """"""

    def save_version(self, version: Version) -> None:
        """

        Parameters
        ----------
        version: Version :


        Returns
        -------

        """
        raise NotImplementedError()


class _ArgumentParserFactoryMixin:
    """"""

    name: str
    description: str

    @classmethod
    def _update_command(cls, sub_parser: ArgumentParser) -> None:
        """

        Parameters
        ----------
        sub_parser: ArgumentParser :


        Returns
        -------

        """
        # Justification: Zen of Python: Explicit is better than implicit
        # Must be implemented if necessary
        pass  # pylint: disable=W0107

    @classmethod
    def get_allowed_arguments(cls) -> set[str]:
        """"""
        return set()

    @classmethod
    def create_command(
        cls, sub_parser_collection: _HasAddSubParser, **kwargs
    ) -> ArgumentParser:
        """

        Parameters
        ----------
        sub_parser_collection: _HasAddSubParser :

        **kwargs :


        Returns
        -------

        """
        aliases: list[str] = []

        if "aliases" in vars(cls):
            aliases = list(getattr(cls, "aliases"))

        exit_on_error = True
        if "exit_on_error" in kwargs:
            exit_on_error = bool(kwargs.pop("exit_on_error"))

        parser = sub_parser_collection.add_parser(
            cls.name,
            description=cls.description,
            aliases=aliases,
            exit_on_error=exit_on_error,
        )

        parser.add_argument(
            "--dry-run",
            "-n",
            action="store_true",
            dest="dry_run",
            help="Do not perform any action, just check if it would work.",
        )

        cls._update_command(parser)

        return parser


class ActionBase(ABC, _ArgumentParserFactoryMixin):
    """"""

    def __init__(self, **kwargs) -> None:
        # Just to ignore key word arguments
        pass

    @abstractmethod
    def run(self, dry_run: bool = False) -> None:
        """

        Parameters
        ----------
        dry_run: bool :
             (Default value = False)

        Returns
        -------

        """
        raise NotImplementedError()

    @classmethod
    def create_from_command(cls, **kwargs) -> "ActionBase":
        """

        Parameters
        ----------
        **kwargs :


        Returns
        -------

        """
        instance: "ActionBase" = cls(**kwargs)

        return instance


class VersionConsumer(ActionBase):
    """"""

    def __init__(self, version: Version, **kwargs) -> None:
        self.__version = version

    @property
    def current_version(self) -> Version:
        """"""
        return self.__version

    @classmethod
    def get_allowed_arguments(cls) -> set[str]:
        """"""
        return set(["version"]).union(ActionBase.get_allowed_arguments())


class VersionModifier(VersionConsumer):
    """"""

    def __init__(
        self, version: Version, persister: VersionPersister, **kwargs
    ) -> None:
        super().__init__(version, **kwargs)
        self.__persister = persister

    @abstractmethod
    def create_new_version(self) -> Version:
        """"""
        raise NotImplementedError()

    def _report_new_version(self, next_version: Version) -> None:
        """

        Parameters
        ----------
        next_version: Version :


        Returns
        -------

        """
        logger.info(
            "Performing increment of version: %s -> %s",
            _formatter.format(self.current_version),
            _formatter.format(next_version),
        )

    def run(self, dry_run: bool = False) -> None:
        """

        Parameters
        ----------
        dry_run: bool :
             (Default value = False)

        Returns
        -------

        """
        next_version: Version = self.create_new_version()
        if not dry_run:
            self.__persister.save_version(next_version)
        else:
            logger.info("Would write new version %s", next_version)

    @classmethod
    def get_allowed_arguments(cls) -> set[str]:
        """"""
        return set(["persister"]).union(
            VersionConsumer.get_allowed_arguments()
        )


class ActionRegistry:
    """"""

    def __init__(self) -> None:
        self.__items: dict[str, type[ActionBase]] = {}

    def register(self) -> Callable:
        """"""

        def decorator(clazz: type[ActionBase]):
            """

            Parameters
            ----------
            clazz: type[ActionBase] :


            Returns
            -------

            """
            if not issubclass(clazz, ActionBase):
                raise ValueError(
                    f"{clazz.__name__} is not an sub-type of "
                    f"{ActionBase.__name__}"
                )
            self.__items[clazz.name] = clazz

            return clazz

        return decorator

    def update_parser(self, parser: ArgumentParser) -> None:
        """

        Parameters
        ----------
        parser: ArgumentParser :


        Returns
        -------

        """
        parsers: _HasAddSubParser = parser.add_subparsers(  # type: ignore
            dest="selected_command",
            description="Either the part of the version to bump "
            + "according to PEP 440:"
            + " major.minor.micro, "
            + "or VCS based actions to take.",
            required=True,
        )
        parser.add_help = True
        keys = list(self.__items.keys())
        keys.sort()
        exit_on_error = False
        if "exit_on_error" in vars(parser):
            exit_on_error = getattr(parser, "exit_on_error")

        for key in keys:
            clazz = self.__items[key]
            clazz.create_command(parsers, exit_on_error=exit_on_error)

    def execute(
        self,
        /,
        args: Namespace,
        *,
        version: Version,
        persister: VersionPersister,
        vcs_provider: VcsProvider,
    ) -> None:
        """

        Parameters
        ----------
        / :

        args: Namespace :

        * :

        version: Version :

        persister: VersionPersister :

        vcs_provider: VcsProvider :


        Returns
        -------

        """
        kwargs: dict = {}

        kwargs.update(vars(args))

        kwargs = {
            key: value
            for key, value in kwargs.items()
            if not key.startswith("_")
        }

        dry_run = False
        if "dry_run" in kwargs:
            dry_run = bool(kwargs.pop("dry_run"))

        selected_command: str = kwargs.pop("selected_command")

        if selected_command not in self.__items:
            raise ValueError(
                f"Failed to get command {selected_command}. "
                + "Valid values are {', '.join(self.__items.keys())}."
            )

        clazz: type[ActionBase] = self.__items[selected_command]

        allowed_args: set[str] = clazz.get_allowed_arguments()
        allowed_kwargs: dict[str, Any] = {
            "version": version,
            "persister": persister,
            "vcs_provider": vcs_provider,
        }

        allowed_kwargs = {
            key: value
            for key, value in allowed_kwargs.items()
            if key in allowed_args
        }

        kwargs.update(allowed_kwargs)

        command: "ActionBase" = clazz.create_from_command(**kwargs)
        command.run(dry_run=dry_run)


actions = ActionRegistry()
action = actions.register()



#
# SPDX-License-Identifier: MIT
#
# Copyright (c) 2021-2023 Carsten Igel.
#
# This file is part of pdm-bump
# (see https://github.com/carstencodes/pdm-bump).
#
# This file is published using the MIT license.
# Refer to LICENSE for more information
#
""""""


from abc import abstractmethod
from argparse import ArgumentParser
from typing import Literal, cast, final

from ..core.logging import logger, traced_function
from ..core.version import NonNegativeInteger, Pep440VersionFormatter, Version
from .base import VersionModifier, VersionPersister, action

_formatter = Pep440VersionFormatter()


class PreviewMismatchError(Exception):
    """"""

    # Justification: Zen of Python: Explicit is better than implicit
    pass  # pylint: disable=W0107


# Justification fulfills a protocol
class _DummyPersister:  # pylint: disable=R0903
    """"""

    def save_version(self, version: Version) -> None:
        """

        Parameters
        ----------
        version: Version :


        Returns
        -------

        """
        # This must not be implemented as it is only a dummy.
        raise NotImplementedError()


class _PreReleaseIncrementingVersionModified(VersionModifier):
    """"""

    def __init__(
        self,
        version: Version,
        persister: VersionPersister,
        increment_micro: bool = True,
        **kwargs,
    ) -> None:
        super().__init__(version, persister, **kwargs)
        self.__increment_micro = increment_micro

    @classmethod
    def _update_command(cls, sub_parser: ArgumentParser) -> None:
        """

        Parameters
        ----------
        sub_parser: ArgumentParser :


        Returns
        -------

        """
        sub_parser.add_argument(
            "--micro",
            action="store_true",
            dest="increment_micro",
            help="When setting pre-release, specifies "
            + "whether micro version shall "
            + "be incremented as well",
        )

        VersionModifier._update_command(sub_parser)

    @traced_function
    def create_new_version(self) -> Version:
        """"""
        letter: Literal["a", "b", "c", "alpha", "beta", "rc"]
        name: str
        letter, name = self.pre_release_part
        pre: tuple[
            Literal["a", "b", "c", "alpha", "beta", "rc"], NonNegativeInteger
        ] = (letter, 1)

        logger.debug(
            "Incrementing %s part of current version %s",
            name,
            _formatter.format(self.current_version),
        )

        if self.current_version.preview is not None:
            if not self._is_valid_preview_version():
                raise PreviewMismatchError(
                    f"{_formatter.format(self.current_version)} "
                    # Weird behavior of sonarlint, pylint and flake8
                    # Variable is declared as unused, if used only in
                    # formatted string
                    + "is not an "
                    + name
                    + " version."
                )
            pre = cast(
                tuple[
                    Literal["a", "b", "c", "alpha", "beta", "rc"],
                    NonNegativeInteger,
                ],
                self.current_version.preview,
            )
            number = pre[1] + 1 if letter == pre[0] else 1

            pre = (letter, number)

        result: Version = Version(
            self.current_version.epoch,
            self._get_next_release(),
            pre,
            None,
            None,
            None,
        )

        self._report_new_version(result)

        return result

    @property
    @abstractmethod
    def pre_release_part(
        self,
    ) -> tuple[Literal["a", "b", "c", "alpha", "beta", "rc"], str]:
        """"""
        raise NotImplementedError()

    @abstractmethod
    def _is_valid_preview_version(self) -> bool:
        """"""
        raise NotImplementedError()

    def _get_next_release(self) -> tuple[NonNegativeInteger, ...]:
        """"""
        micro = self.current_version.micro
        if self.__increment_micro and self.current_version.preview is None:
            micro = micro + 1

        ret: list[NonNegativeInteger] = []
        for val in self.current_version.release:
            ret.append(val)

        while len(ret) < 3:
            ret.append(0)

        ret[2] = micro

        return tuple(ret)


@final
@action
class PreReleaseIncrementingVersionModifier(VersionModifier):
    """"""

    name: str = "pre-release"
    description: str = (
        "Increment a pre-release part (alpha, beta, release-candidate)"
    )

    def __init__(
        self,
        version: Version,
        persister: VersionPersister,
        pre_release_part: str,
        increment_micro: bool = True,
        **kwargs,
    ) -> None:
        super().__init__(version, persister, **kwargs)
        self.__sub_modifier: VersionModifier
        if pre_release_part in (AlphaIncrementingVersionModifier.name,):
            self.__sub_modifier = AlphaIncrementingVersionModifier(
                version, _DummyPersister(), increment_micro
            )
        elif pre_release_part in (BetaIncrementingVersionModifier.name,):
            self.__sub_modifier = BetaIncrementingVersionModifier(
                version, _DummyPersister(), increment_micro
            )
        elif (
            pre_release_part
            in (ReleaseCandidateIncrementingVersionModifier.name,)
            or pre_release_part
            in ReleaseCandidateIncrementingVersionModifier.aliases
        ):
            self.__sub_modifier = ReleaseCandidateIncrementingVersionModifier(
                version, _DummyPersister(), increment_micro
            )
        else:
            raise ValueError(
                f"{pre_release_part} is not a valid pre-release part"
            )

    @traced_function
    def create_new_version(self) -> Version:
        """"""
        return self.__sub_modifier.create_new_version()

    @classmethod
    def _update_command(cls, sub_parser: ArgumentParser) -> None:
        """

        Parameters
        ----------
        sub_parser: ArgumentParser :


        Returns
        -------

        """
        valid_values = []
        valid_values.append(AlphaIncrementingVersionModifier.name)
        valid_values.append(BetaIncrementingVersionModifier.name)
        valid_values.append(ReleaseCandidateIncrementingVersionModifier.name)
        valid_values.extend(
            ReleaseCandidateIncrementingVersionModifier.aliases
        )

        sub_parser.add_argument(
            "--pre",
            action="store",
            type=str,
            default=None,
            choices=valid_values,
            dest="pre_release_part",
            help="Sets a pre-release on the current version."
            + " If a pre-release is set, it can be removed "
            + "using the 'final' option. A new pre-release "
            + "must be greater than the current version."
            + " See PEP440 for details.",
        )

        VersionModifier._update_command(sub_parser)
        # Justification: Class is a mixin
        # pylint: disable=W0212
        _PreReleaseIncrementingVersionModified._update_command(sub_parser)


@final
class AlphaIncrementingVersionModifier(_PreReleaseIncrementingVersionModified):
    """"""

    name: str = "alpha"
    description: str = "Increment the alpha pre-release version part"

    @property
    def pre_release_part(
        self,
    ) -> tuple[Literal["a", "b", "c", "alpha", "beta", "rc"], str]:
        """"""
        return ("a", "alpha")

    def _is_valid_preview_version(self) -> bool:
        """"""
        return self.current_version.is_alpha


@final
class BetaIncrementingVersionModifier(_PreReleaseIncrementingVersionModified):
    """"""

    name: str = "beta"
    description: str = "Increment the beta pre-release version part"

    @property
    def pre_release_part(
        self,
    ) -> tuple[Literal["a", "b", "c", "alpha", "beta", "rc"], str]:
        """"""
        return ("b", "alpha or beta")

    def _is_valid_preview_version(self) -> bool:
        """"""
        return self.current_version.is_alpha or self.current_version.is_beta


@final
class ReleaseCandidateIncrementingVersionModifier(
    _PreReleaseIncrementingVersionModified
):
    """"""

    name: str = "rc"
    description: str = (
        "Increment the release-candidate pre-release version part"
    )
    aliases: tuple[str] = ("c",)

    @property
    def pre_release_part(
        self,
    ) -> tuple[Literal["a", "b", "c", "alpha", "beta", "rc"], str]:
        """"""
        return ("rc", "pre-release")

    def _is_valid_preview_version(self) -> bool:
        """"""
        return (
            self.current_version.is_alpha
            or self.current_version.is_beta
            or self.current_version.is_release_candidate
        )



#
# SPDX-License-Identifier: MIT
#
# Copyright (c) 2021-2023 Carsten Igel.
#
# This file is part of pdm-bump
# (see https://github.com/carstencodes/pdm-bump).
#
# This file is published using the MIT license.
# Refer to LICENSE for more information
#

""""""



#
# SPDX-License-Identifier: MIT
#
# Copyright (c) 2021-2023 Carsten Igel.
#
# This file is part of pdm-bump
# (see https://github.com/carstencodes/pdm-bump).
#
# This file is published using the MIT license.
# Refer to LICENSE for more information
#
""""""


from importlib import import_module
from pathlib import Path
from pkgutil import iter_modules
from traceback import format_exc as get_traceback
from typing import Final

from .logging import logger, traced_function


# Justification: loader should be single instance only.
class _Loader:  # pylint: disable=R0903
    """"""

    def __init__(self) -> None:
        self.__loaded: set[str] = set()

    @traced_function
    def load_modules_of_package(
        self, file_path: str, pkg_name: str, *skip: tuple[str, ...]
    ) -> None:
        """

        Parameters
        ----------
        file_path: str :

        pkg_name: str :

        *skip: tuple[str, ...] :


        Returns
        -------

        """
        file_pth = Path(file_path)
        if not file_pth.exists() or not file_pth.is_file():
            raise FileNotFoundError(file_path)

        pkg_path: Path = file_pth.absolute().parent

        for _, module_name, is_pkg in iter_modules([str(pkg_path)]):
            full_name: str = f"{pkg_name}.{module_name}"
            if (
                module_name not in skip
                and not is_pkg
                and full_name not in self.__loaded
            ):
                try:
                    logger.debug(
                        "Loading module %s from package %s",
                        module_name,
                        pkg_name,
                    )
                    rel_module_name = f".{module_name}"
                    _ = import_module(rel_module_name, pkg_name)
                    self.__loaded.add(full_name)
                except ImportError:
                    logger.exception(
                        "Error loading module %s from %s",
                        module_name,
                        pkg_name,
                        exc_info=False,
                    )
                    logger.debug("Error loading module: %s", get_traceback())


# Justification: Intended to use as a regular variable
loader: Final[_Loader] = _Loader()  # pylint: disable=C0103



#
# SPDX-License-Identifier: MIT
#
# Copyright (c) 2021-2023 Carsten Igel.
#
# This file is part of pdm-bump
# (see https://github.com/carstencodes/pdm-bump).
#
# This file is published using the MIT license.
# Refer to LICENSE for more information
#
""""""


from typing import final

from ..core.logging import logger
from ..core.version import Pep440VersionFormatter, Version
from ..vcs import VcsProvider, VcsProviderAggregator
from .base import VersionConsumer, action


@final
@action
class CreateTagFromVersion(VersionConsumer, VcsProviderAggregator):
    """"""

    name: str = "tag"
    description: str = "Create a VCS revision tag from the current version"

    def __init__(
        self, version: Version, vcs_provider: VcsProvider, **kwargs
    ) -> None:
        VersionConsumer.__init__(self, version, **kwargs)
        VcsProviderAggregator.__init__(self, vcs_provider, **kwargs)

    def run(self, dry_run: bool = False) -> None:
        """

        Parameters
        ----------
        dry_run: bool :
             (Default value = False)

        Returns
        -------

        """
        if not dry_run:
            self.vcs_provider.create_tag_from_version(self.current_version)
        else:
            logger.info(
                "Would create tag v%s",
                Pep440VersionFormatter().format(self.current_version),
            )

    @classmethod
    def get_allowed_arguments(cls) -> set[str]:
        """"""
        return set(["vcs_provider"]).union(
            VersionConsumer.get_allowed_arguments()
        )



#
# SPDX-License-Identifier: MIT
#
# Copyright (c) 2021-2023 Carsten Igel.
#
# This file is part of pdm-bump
# (see https://github.com/carstencodes/pdm-bump).
#
# This file is published using the MIT license.
# Refer to LICENSE for more information
#
""""""


import sys
from abc import abstractmethod
from argparse import ArgumentParser
from dataclasses import asdict as dataclass_to_dict
from typing import Any, Final, final

from ..core.logging import logger, traced_function
from ..core.version import NonNegativeInteger, Pep440VersionFormatter, Version
from .base import VersionModifier, VersionPersister, action

if sys.version_info >= (3, 10, 0):
    # suspicious mypy behavior
    from typing import TypeAlias  # type: ignore
else:
    from typing_extensions import TypeAlias


_formatter = Pep440VersionFormatter()


class _NonFinalPartsRemovingVersionModifier(VersionModifier):
    """"""

    def __init__(
        self,
        version: Version,
        persister: VersionPersister,
        remove_parts: bool = True,
        **kwargs,
    ) -> None:
        super().__init__(version, persister, **kwargs)
        self.__remove_parts = remove_parts

    @property
    def remove_non_final_parts(self) -> bool:
        """"""
        return self.__remove_parts

    @abstractmethod
    def create_new_version(self) -> Version:
        """"""
        raise NotImplementedError()

    @staticmethod
    def _create_new_constructional_args(
        release: tuple[NonNegativeInteger, ...], epoch: NonNegativeInteger = 0
    ) -> dict[str, Any]:
        """

        Parameters
        ----------
        release: tuple[NonNegativeInteger, ...] :

        epoch: NonNegativeInteger :
             (Default value = 0)

        Returns
        -------

        """
        return {
            "epoch": epoch,
            "release_tuple": release,
            "preview": None,
            "post": None,
            "dev": None,
            "local": None,
        }

    @classmethod
    def _update_command(cls, sub_parser: ArgumentParser) -> None:
        """

        Parameters
        ----------
        sub_parser: ArgumentParser :


        Returns
        -------

        """
        sub_parser.add_argument(
            "--no-remove",
            action="store_false",
            dest="remove_parts",
            help="When incrementing major, minor, micro or epoch, "
            + "do not remove all pre-release parts.",
        )

        VersionModifier._update_command(sub_parser)


_NFPR: TypeAlias = _NonFinalPartsRemovingVersionModifier


class _ReleaseVersionModifier(_NonFinalPartsRemovingVersionModifier):
    """"""

    @property
    @abstractmethod
    def release_part(self) -> NonNegativeInteger:
        """"""
        raise NotImplementedError()

    @traced_function
    def create_new_version(self) -> Version:
        """"""
        construction_args: dict[str, Any] = dataclass_to_dict(
            self.current_version
        )

        next_release: tuple[
            NonNegativeInteger, ...
        ] = self._update_release_version_part(self.release_part)

        if self.remove_non_final_parts:
            logger.debug("Removing non-final parts of version")
            # Using type alias due to line length enforced by black
            construction_args = _NFPR._create_new_constructional_args(  # noqa: E501 pylint: disable=W0212
                next_release, self.current_version.epoch
            )
        else:
            construction_args["release_tuple"] = next_release

        next_version: Version = Version(**construction_args)
        self._report_new_version(next_version)

        return next_version

    def _update_release_version_part(
        self, part_id: NonNegativeInteger
    ) -> tuple[NonNegativeInteger, ...]:
        """

        Parameters
        ----------
        part_id: NonNegativeInteger :


        Returns
        -------

        """
        release_part: list[NonNegativeInteger] = list(
            self.current_version.release
        )

        for i in range(len(release_part)):  # pylint: disable=C0200
            logger.debug("Checking if version part at %d must be modified", i)
            if i == part_id:
                logger.debug("Incrementing version part at position %d.", i)
                release_part[i] = release_part[i] + 1
            elif i > part_id:
                logger.debug("Resetting version part at position %d", i)
                release_part[i] = 0

        return tuple(release_part)


@final
@action
class FinalizingVersionModifier(_NonFinalPartsRemovingVersionModifier):
    """"""

    name: str = "no-pre-release"
    description: str = "Remove all pre-release parts from the version"

    def __init__(
        self, version: Version, persister: VersionPersister, **kwargs
    ) -> None:
        super().__init__(version, persister, **kwargs)

    @traced_function
    def create_new_version(self) -> Version:
        """"""
        constructional_args: dict[
            str,
            Any
            # Using type alias due to line length enforced by black
        ] = _NFPR._create_new_constructional_args(  # noqa: E501 pylint: disable=W0212
            self.current_version.release, self.current_version.epoch
        )

        next_version: Version = Version(**constructional_args)
        self._report_new_version(next_version)

        return next_version


@final
@action
class MajorIncrementingVersionModifier(_ReleaseVersionModifier):
    """"""

    name: str = "major"
    description: str = "Increment the major part of the version"

    __MAJOR_PART: Final[NonNegativeInteger] = 0

    @property
    def release_part(self) -> NonNegativeInteger:
        """"""
        return self.__MAJOR_PART


@final
@action
class MinorIncrementingVersionModifier(_ReleaseVersionModifier):
    """"""

    name: str = "minor"
    description: str = "Increment the minor part of the version"

    __MINOR_PART: Final[NonNegativeInteger] = 1

    @property
    def release_part(self) -> NonNegativeInteger:
        """"""
        return self.__MINOR_PART


@final
@action
class MicroIncrementingVersionModifier(_ReleaseVersionModifier):
    """"""

    name: str = "micro"
    description: str = "Increment the micro (or patch) part of the version"
    aliases: tuple[str] = ("patch",)

    __MICRO_PART: Final[NonNegativeInteger] = 2

    @property
    def release_part(self) -> int:
        """"""
        return self.__MICRO_PART


@final
@action
class EpochIncrementingVersionModifier(_NonFinalPartsRemovingVersionModifier):
    """"""

    name: str = "epoch"
    description: str = "Increment the epoch of the version"

    def __init__(
        self,
        version: Version,
        persister: VersionPersister,
        remove_parts: bool = True,
        reset_version: bool = True,
    ):
        super().__init__(version, persister, remove_parts)
        self.__reset_version = reset_version

    @traced_function
    def create_new_version(self) -> Version:
        """"""
        constructional_args: dict[str, Any] = dataclass_to_dict(
            self.current_version
        )

        if self.__reset_version or self.remove_non_final_parts:
            constructional_args = dataclass_to_dict(Version.default())
            if not self.__reset_version:
                logger.debug("Current version tuple shall not be reset")
                constructional_args[
                    "release_tuple"
                ] = self.current_version.release

        logger.debug("Incrementing Epoch of version")
        constructional_args["epoch"] = self.current_version.epoch + 1

        next_version: Version = Version(**constructional_args)
        self._report_new_version(next_version)
        return next_version

    @classmethod
    def _update_command(cls, sub_parser: ArgumentParser) -> None:
        """

        Parameters
        ----------
        sub_parser: ArgumentParser :


        Returns
        -------

        """
        sub_parser.add_argument(
            "--reset",
            action="store_true",
            dest="reset_version",
            help="Reset version to 1.0.0 on epoch increment",
        )

        # Justification: Call protected method of parent class
        _NFPR._update_command(sub_parser)  # pylint: disable=W0212


@final
@action
class DevelopmentVersionIncrementingVersionModifier(VersionModifier):
    """"""

    name: str = "dev"
    description: str = "Increment the local development part"

    @traced_function
    def create_new_version(self) -> Version:
        """"""
        dev_version: NonNegativeInteger = 1
        if self.current_version.dev is not None:
            _, dev_version = self.current_version.dev
            logger.debug("Incrementing development version part by one")
            dev_version = dev_version + 1

        constructional_args: dict[str, Any] = dataclass_to_dict(
            self.current_version
        )
        constructional_args["dev"] = ("dev", dev_version)

        next_version: Version = Version(**constructional_args)
        self._report_new_version(next_version)

        return next_version


@final
@action
class PostVersionIncrementingVersionModifier(VersionModifier):
    """"""

    name: str = "post"
    description: str = "Increment the post version part"

    @traced_function
    def create_new_version(self) -> Version:
        """"""
        post_version: NonNegativeInteger = 1
        if self.current_version.post is not None:
            _, post_version = self.current_version.post
            logger.debug("Incrementing post version part by one")
            post_version = post_version + 1

        constructional_args: dict[str, Any] = dataclass_to_dict(
            self.current_version
        )
        constructional_args["post"] = ("post", post_version)

        next_version: Version = Version(**constructional_args)
        self._report_new_version(next_version)

        return next_version



#
# SPDX-License-Identifier: MIT
#
# Copyright (c) 2021-2023 Carsten Igel.
#
# This file is part of pdm-bump
# (see https://github.com/carstencodes/pdm-bump).
#
# This file is published using the MIT license.
# Refer to LICENSE for more information
#
""""""


from collections.abc import Mapping
from logging import (
    CRITICAL,
    DEBUG,
    ERROR,
    INFO,
    NOTSET,
    WARN,
    WARNING,
    Filter,
    Handler,
    Logger,
    LogRecord,
    StreamHandler,
    getLogger,
)
from sys import stderr, stdout
from typing import Any, Final, Optional, cast

# MyPy does not recognize this during pull requests
from pdm.termui import UI, Verbosity  # type: ignore


def _get_has_rich():
    """"""
    try:
        # Justification: Required to find module, not more
        import rich  # noqa: F401 pylint: disable=C0415,W0611
    except ImportError:
        return False

    return True


HAS_RICH: Final[bool] = _get_has_rich()

if HAS_RICH:
    from rich.console import Console
    from rich.default_styles import DEFAULT_STYLES
    from rich.logging import RichHandler
    from rich.theme import Style, StyleType, Theme


TRACE: Final[int] = 5


class TermUIHandler(Handler):
    """"""

    def __init__(self, ui: UI, level: int = NOTSET) -> None:
        super().__init__(level)
        self.__ui = ui

    def emit(self, record: LogRecord) -> None:
        """

        Parameters
        ----------
        record: LogRecord :


        Returns
        -------

        """
        msg_format = "{text}"
        if record.levelno in (WARN, WARNING):
            msg_format = "[warning]{text}[/warning]"
        verbosity: Verbosity = Verbosity.NORMAL
        if record.levelno == DEBUG:
            verbosity = Verbosity.DETAIL
        if record.levelno == TRACE:
            verbosity = Verbosity.DEBUG

        text = msg_format.format(text=self.format(record))
        self.__ui.echo(
            text,
            record.levelno in (ERROR, CRITICAL),
            verbosity,
        )


# Justification: Only one method to override
class _ErrorWarningsFilter(Filter):  # pylint: disable=R0903
    """"""

    def __init__(self, invert: Optional[bool] = False) -> None:
        self.__invert: bool = invert or False
        super().__init__()

    def filter(self, record: LogRecord) -> bool:
        """

        Parameters
        ----------
        record: LogRecord :


        Returns
        -------

        """
        warning_and_above: tuple[int, ...] = (WARN, WARNING, ERROR, CRITICAL)
        return (
            record.levelno in warning_and_above
            if not self.__invert
            else record.levelno not in warning_and_above
        )


class TracingLogger(Logger):
    """"""

    def trace(self, msg: str, *args: tuple[Any, ...], **kwargs) -> None:
        """

        Parameters
        ----------
        msg: str :

        *args: tuple[Any, ...] :

        **kwargs :


        Returns
        -------

        """
        self.log(TRACE, msg, *args, **kwargs)

    # Justification: Overriding inherited method
    def makeRecord(  # noqa: N802 pylint: disable=R0913
        self,
        name: str,
        level: int,
        fn: str,
        lno: int,
        msg: Any,
        args: Any,
        exc_info: Any,
        func: Optional[str] = None,
        extra: Optional[Mapping[str, Any]] = None,
        sinfo: Optional[str] = None,
    ) -> LogRecord:
        """

        Parameters
        ----------
        name: str :

        level: int :

        fn: str :

        lno: int :

        msg: Any :

        args: Any :

        exc_info: Any :

        func: Optional[str] :
             (Default value = None)
        extra: Optional[Mapping[str :

        Any]] :
             (Default value = None)
        sinfo: Optional[str] :
             (Default value = None)

        Returns
        -------

        """
        record: LogRecord = super().makeRecord(
            name, level, fn, lno, msg, args, exc_info, func, extra, sinfo
        )
        if level == TRACE:
            record.levelname = "TRACE"

        return record


Logger.manager.setLoggerClass(TracingLogger)


def _get_rich_logger() -> TracingLogger:
    """"""
    _logger: TracingLogger = cast(TracingLogger, getLogger("pdm-bump"))

    styles: dict[str, StyleType] = {}
    styles.update(DEFAULT_STYLES)
    styles.update(
        {
            "logging.level.notset": Style(dim=True),
            "logging.level.trace": Style(dim=True, color="bright_black"),
            "logging.level.debug": Style(dim=True, color="white"),
            "logging.level.info": Style(color="blue"),
            "logging.level.warning": Style(color="yellow"),
            "logging.level.error": Style(color="bright_red"),
            "logging.level.critical": Style(color="red"),
        }
    )
    logging_theme: Theme = Theme(styles)

    std_out: Console = Console(stderr=False, theme=logging_theme)
    std_err: Console = Console(stderr=True, theme=logging_theme)

    std_out_handler: Handler = RichHandler(console=std_out)
    std_err_handler: Handler = RichHandler(console=std_err)

    std_out_handler.addFilter(_ErrorWarningsFilter(True))
    std_err_handler.addFilter(_ErrorWarningsFilter(False))

    _logger.addHandler(std_out_handler)
    _logger.addHandler(std_err_handler)

    _logger.level = INFO

    return _logger


def _get_std_logger() -> TracingLogger:
    """"""
    _logger: TracingLogger = cast(TracingLogger, getLogger("pdm-bump"))
    # mypy: No overload variant of "StreamHandler"
    #       matches argument type "Handler"
    std_out: Handler = StreamHandler(stream=stdout)  # type: ignore
    std_err: Handler = StreamHandler(stream=stderr)  # type: ignore

    std_out.addFilter(_ErrorWarningsFilter(True))
    std_err.addFilter(_ErrorWarningsFilter(False))

    _logger.addHandler(std_out)
    _logger.addHandler(std_err)

    _logger.level = INFO

    return _logger


# Justification: Intended to use as a regular variable
logger: Final[TracingLogger] = (  # pylint: disable=C0103
    _get_std_logger() if not HAS_RICH else _get_rich_logger()
)


def update_logger_from_project_ui(ui_instance: UI) -> None:
    """

    Parameters
    ----------
    ui_instance: UI :


    Returns
    -------

    """
    for handler in list(logger.handlers):
        logger.removeHandler(handler)

    new_handler: Handler = TermUIHandler(ui_instance)
    logger.addHandler(new_handler)


def setup_logger(level: int) -> None:
    """

    Parameters
    ----------
    level: int :


    Returns
    -------

    """
    if 0 > level:
        return

    if level == 0:
        logger.setLevel(INFO)
    elif level == 1:
        logger.setLevel(DEBUG)
    elif level == 2:
        logger.setLevel(TRACE)


def traced_function(fun):
    """

    Parameters
    ----------
    fun :


    Returns
    -------

    """

    def tracing_function(*args: tuple[Any, ...], **kwargs):
        """

        Parameters
        ----------
        *args: tuple[Any, ...] :

        **kwargs :


        Returns
        -------

        """
        try:
            logger.trace(
                "%s: Entering function", fun.__qualname__ or fun.__name__
            )
            return fun(*args, **kwargs)
        finally:
            logger.trace(
                "%s: Exiting function", fun.__qualname__ or fun.__name__
            )

    return tracing_function



#
# SPDX-License-Identifier: MIT
#
# Copyright (c) 2021-2023 Carsten Igel.
#
# This file is part of pdm-bump
# (see https://github.com/carstencodes/pdm-bump).
#
# This file is published using the MIT license.
# Refer to LICENSE for more information
#
""""""


from abc import ABC, abstractmethod
from collections.abc import Iterator
from pathlib import Path
from typing import AnyStr, Callable, Final, NamedTuple, Optional, Union, cast

from ..core.version import Pep440VersionFormatter, Version

_Pathlike = Union[Path, AnyStr]


class _PathLikeConverter(ABC):
    """"""

    @staticmethod
    def _pathlike_to_path(path: _Pathlike) -> Path:
        """

        Parameters
        ----------
        path: _Pathlike :


        Returns
        -------

        """
        if isinstance(path, Path):
            return cast(Path, path)
        if isinstance(path, str):
            return Path(path)
        if isinstance(path, bytes):
            return Path(cast(bytes, path).decode("utf-8"))

        raise ValueError(
            f"'{path}' must be a valid path, string or bytes instance"
        )


class VcsFileSystemIdentifier(NamedTuple):
    """"""

    file_name: Optional[str]
    dir_name: Optional[str]


class VcsProvider(_PathLikeConverter, ABC):
    """"""

    def __init__(self, path: _Pathlike) -> None:
        self.__path = _PathLikeConverter._pathlike_to_path(path)

    @property
    def is_available(self) -> bool:
        """"""
        return False

    @property
    def current_path(self) -> Path:
        """"""
        return self.__path

    @property
    @abstractmethod
    def is_clean(self) -> bool:
        """"""
        raise NotImplementedError()

    @abstractmethod
    def check_in_items(self, message: str, *files: tuple[Path, ...]) -> None:
        """

        Parameters
        ----------
        message: str :

        *files: tuple[Path, ...] :


        Returns
        -------

        """
        raise NotImplementedError()

    def create_tag_from_version(
        self, version: Version, prepend_letter_v: bool = True
    ) -> None:
        """

        Parameters
        ----------
        version: Version :

        prepend_letter_v: bool :
             (Default value = True)

        Returns
        -------

        """
        version_formatted: str = Pep440VersionFormatter().format(version)
        if prepend_letter_v:
            version_formatted = "v" + version_formatted
        self.create_tag_from_string(version_formatted)

    @abstractmethod
    def create_tag_from_string(self, version_formatted: str) -> None:
        """

        Parameters
        ----------
        version_formatted: str :


        Returns
        -------

        """
        raise NotImplementedError()

    @abstractmethod
    def get_most_recent_tag(self) -> Optional[Version]:
        """"""
        raise NotImplementedError()

    @abstractmethod
    def get_number_of_changes_since_last_release(self) -> int:
        """"""
        raise NotImplementedError()

    @abstractmethod
    def get_changes_not_checked_in(self) -> int:
        """"""
        raise NotImplementedError()


class VcsProviderAggregator:
    """"""

    def __init__(self, vcs_provider: VcsProvider, **kwargs) -> None:
        self.__vcs_provider: VcsProvider = vcs_provider

    @property
    def vcs_provider(self) -> VcsProvider:
        """"""
        return self.__vcs_provider


class VcsProviderFactory(_PathLikeConverter, ABC):
    """"""

    def force_create_provider(self, path: Path) -> VcsProvider:
        """

        Parameters
        ----------
        path: Path :


        Returns
        -------

        """
        return self._create_provider(path)

    @abstractmethod
    def _create_provider(self, path: Path) -> VcsProvider:
        """

        Parameters
        ----------
        path: Path :


        Returns
        -------

        """
        raise NotImplementedError()

    @property
    @abstractmethod
    def vcs_fs_root(self) -> Iterator[VcsFileSystemIdentifier]:
        """"""
        raise NotImplementedError()

    def find_repository_root(self, path: _Pathlike) -> Optional[VcsProvider]:
        """

        Parameters
        ----------
        path: _Pathlike :


        Returns
        -------

        """
        real_path: Path = _PathLikeConverter._pathlike_to_path(path)
        return self.find_repository_root_from_path(real_path)

    def find_repository_root_from_path(
        self, path: Path
    ) -> Optional[VcsProvider]:
        """

        Parameters
        ----------
        path: Path :


        Returns
        -------

        """
        if not path.is_dir():
            raise ValueError(f"{path} must refer to a directory.")

        cur_path: Path = path
        while len(cur_path.parts) > 1:  # First part refers to root level
            if self._is_valid_root(cur_path):
                return self._create_provider(cur_path)
            cur_path = cur_path.parent

        return None

    def _is_valid_fs_root_file(self, path: Path) -> bool:
        """

        Parameters
        ----------
        path: Path :


        Returns
        -------

        """
        return path is not None and path.exists() and path.is_file()

    def _is_valid_fs_root_dir(self, path: Path) -> bool:
        """

        Parameters
        ----------
        path: Path :


        Returns
        -------

        """
        return path is not None and path.exists() and path.is_dir()

    def _directory_exists(self, path: Path, directory_name: str) -> bool:
        """

        Parameters
        ----------
        path: Path :

        directory_name: str :


        Returns
        -------

        """
        dir_path: Path = path / directory_name
        return self._is_valid_fs_root_dir(dir_path)

    def _file_exists(self, path: Path, file_name: str) -> bool:
        """

        Parameters
        ----------
        path: Path :

        file_name: str :


        Returns
        -------

        """
        file_path: Path = path / file_name
        return self._is_valid_fs_root_file(file_path)

    def _is_valid_root(self, path: Path) -> bool:
        """

        Parameters
        ----------
        path: Path :


        Returns
        -------

        """
        is_repository_root = False
        for fs_root in self.vcs_fs_root:
            is_valid_root_by_file: bool = (
                fs_root.file_name is None
                or self._file_exists(path, fs_root.file_name)
            )
            is_valid_root_by_dir: bool = (
                fs_root.dir_name is None
                or self._directory_exists(path, fs_root.dir_name)
            )
            is_repository_root = is_repository_root or (
                is_valid_root_by_dir and is_valid_root_by_file
            )

        return is_repository_root


class VcsProviderRegistry(
    dict[str, Callable[..., VcsProviderFactory]], _PathLikeConverter
):
    """"""

    def find_repository_root(self, path: _Pathlike) -> Optional[VcsProvider]:
        """

        Parameters
        ----------
        path: _Pathlike :


        Returns
        -------

        """
        real_path: Path = _PathLikeConverter._pathlike_to_path(path)
        return self.find_repository_root_by_path(real_path)

    def find_repository_root_by_path(
        self, path: Path
    ) -> Optional[VcsProvider]:
        """

        Parameters
        ----------
        path: Path :


        Returns
        -------

        """
        for _, value in self.items():
            factory: VcsProviderFactory = value()
            result: Optional[
                VcsProvider
            ] = factory.find_repository_root_from_path(path)
            if result is not None:
                return result

        return None

    def register(self, name: str) -> Callable:
        """

        Parameters
        ----------
        name: str :


        Returns
        -------

        """

        def decorator(clazz: type[VcsProvider]):
            """

            Parameters
            ----------
            clazz: type[VcsProvider] :


            Returns
            -------

            """
            if not issubclass(clazz, VcsProviderFactory):
                raise ValueError(
                    f"{clazz.__name__} is not an sub-type of "
                    f"{VcsProviderFactory.__name__}"
                )

            self[name] = clazz

        return decorator

    def __missing__(
        self, key: str
    ) -> Optional[Callable[..., VcsProviderFactory]]:
        return None


vcs_providers: Final[VcsProviderRegistry] = VcsProviderRegistry()

vcs_provider: Final[Callable[[str], Callable]] = vcs_providers.register


class DefaultVcsProvider(VcsProvider):
    """"""

    @property
    def is_available(self) -> bool:
        """"""
        return True

    @property
    def is_clean(self) -> bool:
        """"""
        return True

    def check_in_items(self, message: str, *files: tuple[Path, ...]) -> None:
        """

        Parameters
        ----------
        message: str :

        *files: tuple[Path, ...] :


        Returns
        -------

        """
        # Must not be provided
        pass

    def create_tag_from_string(self, version_formatted: str) -> None:
        """

        Parameters
        ----------
        version_formatted: str :


        Returns
        -------

        """
        # Must not be provided
        pass

    def get_most_recent_tag(self) -> Optional[Version]:
        """"""
        # Cannot be provided
        return None

    def get_number_of_changes_since_last_release(self) -> int:
        """"""
        # Cannot be provided
        return 0

    def get_changes_not_checked_in(self) -> int:
        """"""
        # Cannot be provided
        return 0


class VcsProviderError(Exception):
    """"""

    pass
